/***************************************************************************
*   Copyright (C) 2003 by Hideki Ikemoto                                  *
*   ikemo@users.sourceforge.jp                                            *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
***************************************************************************/

#include "kitathreadtabwidget.h"

#include "kitathreadpart.h"
#include "kitathreadview.h"

#include <klibloader.h>
#include <kpopupmenu.h>
#include <krun.h>
#include <kdebug.h>

#include <qclipboard.h>
#include <qapplication.h>
#include <qmessagebox.h>
#include <qregexp.h>

#include <kxmlguifactory.h>
#include <kxmlguiclient.h>
#include <kaction.h>
#include <kstdaccel.h>
#include <kstdaction.h>

#include "libkita/kita_misc.h"
#include "libkita/parsemisc.h"
#include "libkita/signalcollection.h"

#define MAX_TABLABEL_LEN 20

enum {
    ID_Browser_Open,
    ID_Copy_Title_URL,
    ID_Close
};

KitaThreadTabWidget::KitaThreadTabWidget( QWidget* parent, const char* name, WFlags f )
        : QTabWidget( parent, name, f )
{
    setTabBar( new KitaThreadTabBar( this ) );
    connect( tabBar(), SIGNAL( deleteMenuSelected( KitaThreadView* ) ),
             SLOT( deleteView( KitaThreadView* ) ) );

    KGlobal::locale() ->insertCatalogue( "kitapart" );
    KitaThreadView* view = createView();

    if ( view ) {
        m_viewList.append( view );
        addTab( view, "thread" );
        connectSignals( view );
    }
    connect( this, SIGNAL( currentChanged ( QWidget * ) ),
             SLOT( slotCurrentChanged ( QWidget * ) ) );

    setupActions();
}

KitaThreadTabWidget::~KitaThreadTabWidget()
{}


/*-------------------------------*/
/* setup actions                 */  
/*
   Child's actions are added in
   slotCurrentChanged later.     */  /* private */
void KitaThreadTabWidget::setupActions(){

    setXMLFile("kitathreadtabwidgetui.rc");
    
    new KAction( i18n( "Close Current Thread Tab" ),
                 KStdAccel::close(),
                 this,
                 SLOT( closeCurrentTab() ),
                 actionCollection(),
                 "threadtab_closetab" );
    
    m_oldXmlClient = NULL;
}


void KitaThreadTabWidget::showThread( const QString& datURL, bool withNewTab )
{
    QString refstr;
    QString urlstr = Kita::ParseMisc::parseURL(datURL,refstr).prettyURL();
    Kita::Thread * thread = Kita::Thread::getByURL( urlstr );

    emit sigHideSubject();

    KitaThreadView * view = findView( thread->url() );
    if ( view ) {
        setCurrentPage( indexOf( view ) );
        if ( view->threadURL().isEmpty() ) {
            view->showThread( thread );
        } else {
            view->slotReloadButton();
        }
    } else if ( withNewTab ) {
        KitaThreadView * newView = createView();

        if ( newView ) {
            addTab( newView, Kita::unescape( thread->name().left( MAX_TABLABEL_LEN ) ) );
            connectSignals( newView );
            newView->showThread( thread );
            showPage( newView );
            m_viewList.append( newView );
        }
    } else {
        static_cast<KitaThreadView *>( currentPage() ) ->showThread( thread );
    }

    setTabLabel( currentPage(), Kita::unescape( thread->name().left( MAX_TABLABEL_LEN ) )
                 + QString( " (%1)" ).arg( thread->resNum() ) );
    setTabToolTip( currentPage(), thread->name() );
}

KitaThreadView* KitaThreadTabWidget::createView()
{
    KLibFactory * factory = KLibLoader::self() ->factory( "libkitapart" );
    if ( ! factory ) {
        QMessageBox::critical( parentWidget(),
                               i18n( "Load Error" ),
                               i18n( "can't load libkitapart." ) );
        return 0;
    }
    KitaThreadPart* part = static_cast<KitaThreadPart *>( factory->create( this, "thread", "KitaThreadPart" ) );
    KitaThreadView* view = static_cast<KitaThreadView *>( part->widget() );

    return view;
}

void KitaThreadTabWidget::connectSignals( KitaThreadView* view )
{
    Kita::SignalCollection* signalCollection = Kita::SignalCollection::getInstance();
    connect( this, SIGNAL( signalChangeStatusbar( const QString& ) ),
             signalCollection, SIGNAL( signalChangeStatusbar( const QString& ) ) );
    connect( signalCollection, SIGNAL( signalThread( const Kita::Thread* ) ),
             this, SLOT( slotThread( const Kita::Thread* ) ) );
}

KitaThreadView* KitaThreadTabWidget::findView( const QString& threadURL )
{
    KitaThreadView * view;
    for ( view = m_viewList.first(); view; view = m_viewList.next() ) {
        if ( view->threadURL().url() == threadURL ) {
            return view;
        }
        if ( view->threadURL().isEmpty() ) {
            return view; // default view
        }
    }
    return 0;
}

void KitaThreadTabWidget::deleteView( KitaThreadView* view )
{
    kdDebug() << "deleteView(" << static_cast<void*>( view ) << ")" << endl;

    factory()->removeClient( view );
    m_oldXmlClient = NULL;

    removePage( view );
    m_viewList.remove( view );
    delete view;

    if ( count() == 0 ) {
        KitaThreadView * view = createView();

        if ( view ) {
            m_viewList.append( view );
            addTab( view, "thread" );
            connectSignals( view );

            showPage( view );
            emit signalChangeStatusbar( "" );
            topLevelWidget() ->setCaption( "" );
        }
    }
}

void KitaThreadTabWidget::focusSearchCombo()
{
    static_cast<KitaThreadView *>( currentPage() ) ->focusSearchCombo();
}

void KitaThreadTabWidget::reloadThread()
{
    static_cast<KitaThreadView *>( currentPage() ) ->slotReloadButton();
}

void KitaThreadTabWidget::closeCurrentTab()
{
    deleteView( static_cast<KitaThreadView *>( currentPage() ) );
}

KitaThreadTabBar::KitaThreadTabBar( QWidget* parent, const char* name )
        : QTabBar( parent, name )
{}

KitaThreadTabBar::~KitaThreadTabBar()
{}

void KitaThreadTabBar::contextMenuEvent( QContextMenuEvent* e )
{
    KPopupMenu popup( 0 );
    popup.insertItem( i18n( "Open with Web Browser" ), ID_Browser_Open );
    popup.insertItem( i18n( "Copy title and URL" ), ID_Copy_Title_URL );
    popup.insertItem( i18n( "Close this tab" ), ID_Close );

    QTabWidget* parent = static_cast<QTabWidget*>( parentWidget() );
    KitaThreadView* threadView = static_cast<KitaThreadView*>( parent->currentPage() );
    QClipboard* clipboard = QApplication::clipboard();

    switch ( popup.exec( e->globalPos() ) ) {
    case ID_Browser_Open:
        KRun::runURL( threadView->threadURL(), "text/html" );
        break;
    case ID_Copy_Title_URL:
        clipboard->setText( threadView->threadName() + "\n" + threadView->threadURL().url() );
        break;
    case ID_Close:
        emit deleteMenuSelected( threadView );
        break;
    default:
        break;
    }

    kdDebug() << "currentPage = " << tab( currentTab() ) ->text() << endl;
}

void KitaThreadTabWidget::slotCurrentChanged( QWidget * w )
{
    KitaThreadView* view = static_cast<KitaThreadView *>( w );
    if(m_oldXmlClient) factory()->removeClient(m_oldXmlClient);
    factory()->addClient(view);
    view->showStatusBar( "" );
    m_oldXmlClient = view;
}

void KitaThreadTabWidget::slotThread( const Kita::Thread* thread )
{
    KitaThreadView * view = findView( thread->url() );
    if ( view ) {
        setTabLabel( currentPage(), Kita::unescape( thread->name().left( MAX_TABLABEL_LEN ) ) + QString( " (%1)" ).arg( thread->resNum() ) );
    }
}
